=begin
@author		Ville Kumpulainen

A class implementing DBSCAN.
=end

require 'dataset'
require 'metric'
require 'clustering'

class DBSCAN < Clustering
	attr_reader :threshold, :k, :splits, :sections

	#
	# Initializer.
	#
	#@param	data	Dataset representing data vectors.
	#@param	p		Possible initial parameters of DBSCAN as a Hash. Keywords 'Eps' and 'MinPts' (and their alternatives 'threshold' and 'k') affecting the clustering and 'splits' and 'sections' affecting threshold deduction are supported.
	#
	def initialize data, p = {}
		super(data)
		@threshold = nil
		@k = 4
		@splits = 3
		@sections = nil
		
		p.each_pair { |k,v|
			case k
				when %r{^(Eps|threshold)$}
					raise "Parameter Eps must be a positive Fixnum/Float" unless (v.class == Fixnum || v.class == Float) && v > 0
					@threshold = v
				when %r{^(MinPts|k)$}
					raise "Parameter k must be a positive Fixnum" unless v.class == Fixnum && v > 0
					@k = v
				when %r{^(splits)$}
					raise "Parameter splits must be a Fixnum with at least a value of 3" unless v.class == Fixnum && v >= 3
					@splits = v
				when %r{^(sections)$}
					raise "Parameter sections must be a positive Fixnum" unless v.class == Fixnum && v > 0
					@sections = v
				else
					raise "Undetected parameter #{k}"
			end
		}
		
		@sections = 2 ** Math.log(@splits - 1, 2).to_i if @sections.nil? # Default behaviour: take at most half off on every iteration. O(n).
#		@sections = 1 unless @sections.nil? # Another proposal for default behaviour: with @splits being more than 2 time complexity is O(log_{@splits} n), but the algorithm can easily fixate to local maximums.
		raise "There must be less sections than splits" if @sections >= @splits
		
		#@todo: Utilize Dataset (most preferably one using a database) to store dist? It takes only O(n) memory, but on the other hand, we're prepared to store some other O(n) data into a Dataset as well, just in case that O(n) really is too much to handle in a simple array..
		if @threshold.nil?
			raise "Not enough data for finding out #{@k}-dists!" if @data.length < @k
			dist = [] #map k-dists to this array
			i = 0
			for i in 0...@data.length
				puts "#{i}\t/ #{@data.length}" if i%100 == 0
				nn = @data.nearestNeighbors i, @k
				dist << Metric::distance(@data[nn.last], @data[i])
			end
			puts "Sorting.."
			dist.sort!
			puts "sorting finished"

			range = 0..(dist.length - 1)
			
			while(range.count > @splits + 1) do
				val = []
				for i in 0..@splits
					val << (range.begin + (range.end - range.begin) * i / @splits).to_i
				end
				
				step = (range.end - range.begin) / @splits
				last_angles = [Math.atan((dist[val[1]] - dist[val[0]]) / step)]
#				last_angles = [Math.atan((dist[val[@sections]] - dist[val[@sections - 1]]) / step)] * 2
				best_skew = 0
				
				c = 0
				for i in 2..@splits
#				for i in ((@sections) / 2)..(@splits - (@sections) / 2)
					last_angles[(i - 1) % 2] = Math.atan((dist[val[i]] - dist[val[i - 1]]) / step)
					if (current_skew = (last_angles[0] - last_angles[1]).abs) > best_skew
						c += 1
						best_skew = current_skew
#						range = val[i - 2]..val[i]
#						range = val[i - (@sections + 1) / 2]..val[i + @sections / 2]
						range = val[(i - 1 - (@sections + 1) / 2 < 0 ? 0 : i - 1 - (@sections + 1) / 2)]..val[(i - 1 + (@sections + 1) / 2 > val.length - 1 ? val.length - 1 : i - 1 + (@sections + 1) / 2)]
					end
				end
				puts "\t#{@sections} / #{@splits}\t#{range.inspect}\t#{c}"
			end

			@threshold = dist[range].inject(0){ |s,i| s += i } / range.count #pick average

			puts "Threshold set to #{@threshold}"
		end
		@i = 0
		@cid = 1
	end
	
	def step
		cid = 1
		
		#if flatten to steps of iterations, these variables must be stored between calls: i, cid (= @data.clusters?)
		
#		for i in 0...@data.length
#			puts "#{i+1}\t/\t#{@data.length} (cid: #{cid})" if i < 25 or i%100 == 0
			labeled = []
			if @data.getLabel(@i).nil? || @data.getLabel(@i) == 0 #unclassified / noise
				seeds = [@i]
				while(s = seeds.pop) do
					neighbors = @data.regionalQuery s, @threshold
					
					if neighbors.length >= @k #s is a core point
						(@data.setLabel(s, -@cid) ; labeled << s) if @data.getLabel(s).nil? || @data.getLabel(s) == 0 #label as 'gray', if 'white' or earlier labeled as noise (re-labeling 'gray' would result repetition, re-labeling 'black' is against definition of dbscan)
						neighbors.each { |n|
							if @data.getLabel(n).nil? || @data.getLabel(n) == 0
								seeds << n
								@data.setLabel(n, -@cid) ; labeled << n
							end
							# @data.label(n) < 0: already visited and labeled as cid
							# @data.label(n) > 0: already visited, labeled as <cid and by definition not re-labeled
						}
					elsif s != @i #initial point i was a core point and s is a boundary point density-reachable from i
						(@data.setLabel(s, -@cid) ; labeled << s) if @data.getLabel(s).nil? || @data.getLabel(s) == 0
					else
						@data.setLabel(s, 0) #classify as noise; might be a boundary point found later as well..
					end
				end
			end
			labeled.each { |m| @data.setLabel(m, @data.getLabel(m).abs) } #depth-first search finished, re-label 'gray' as 'black'
			@data.clusters = (@cid += 1) - 1 if not labeled.empty? #neighboring points were visited => cluster id was used and should be changed
#		end
		
		@i += 1
		
#		@data.clusters = cid
		
#		puts "Finished!"
	end
	
	def happy?
		not @i < @data.length
	end
end
