=begin
@author Ville Kumpulainen

An "abstract" class describing a common interface to access data. This class should serve as a starting point for storaging-specific implementations (such as plain array, R*-tree, temporary file or a database). Methods accessing data should - for obvious reasons - remain the same in derived classes, but methods adding data to the Dataset may vary.

Data in this context consists of data vectors and their labels (representing partitioning). Representative centroids and distances may also be stored, if provided by the clustering algorithm.

@todo: There are not any handy methods for K-means (distances, partitioning).
@todo: Rename some of the methods; method_name seems to be a habit in Ruby..
@todo: Freeze data before clustering. Data must not be modified after the clustering process has started.
=end

class Dataset
	#
	# A simple method for adding value in the Dataset.
	#
	#@param	value	Array of coordinates of the point to add.
	#
	def add value
		raise "Not implemented"
	end
	
	#
	# A simple method for adding value in the Dataset. Calls add with provided parameter.
	#
	#@param	value	Array of coordinates of the point to add.
	#@see	add
	#
	def << value
		add value
	end
	
	#
	# Dimensionality of the stored data.
	#
	#@return	Dimensionality.
	#
	def dimensionality
		raise "Not implemented"
	end
	
	def dim #shortened form
		dimensionality
	end
	
	#
	# Bounding box of the dataset.
	#
	#@return	Array of two cell objects, each representing minimum and maximum coordinate values on a single axis.
	#
	def boundaries
		raise "Not implemented"
	end
	
	#
	# Number of vectors in a dataset.
	#
	#@return	Number of vectors in a dataset.
	#
	def length
		raise "Not implemented"
	end
	
	#
	# A simple method to access value in Dataset.
	#
	#@param	i	Index of desired data point.
	#@return	Array consisting of coordinates of the desired data point.
	#
	def [] i
		raise "Not implemented"
	end

	#
	# Query all points near (i)th point. If a block is given, r is discarded and the block is used as a criterion to decide whether or not to consider point as a neighbor. Parameters provided for the optional block are coordinates for original point and its neighbor candidate.
	#
	#@param	i	Center point for query.
	#@param	r	Radius for query.
	#@return	Indexes of points having at most distance of r to (i)th point.
	#
	def regionalQuery i, r
		raise "Not implemented"
	end
	
	def regionalNeighbors i, r
		regionalQuery i, r
	end
	
	#
	# Queries k nearest points surrounding (i)th point.
	#
	#@param	i	Index of a point.
	#@param	k	Number of nearest neighbors to return.
	#@return	Indexes of k closest points to (i)th point.
	#
	def nearestNeighbors i, k
		raise "Not implemented"
	end
	
	#@todo: Implemented setters should freeze data modification.
	
	#
	# Get label(s) of given data point(s).
	#
	#@param	i	Index of data point or indexes described by an Array or Range.
	#
	def getLabel i
		raise "Not implemented"
	end
	
	#
	# Set label of (i)th data point. Labels should refer to centroid indexes.
	#
	#@param	i	Index of data point to be labeled.
	#@param	l	Label (or nil).
	#
	def setLabel i, l
		raise "Not implemented"
	end
	
	#
	# Returns points labeled as l or matching to given block.
	#
	#@param	label	Label.
	#@yield			Find all points whose label results true on a given block.
	#@return		Array of data point indexes labeled as l or having label matching to the provided block.
	#
	def get_points_labeled label = 0
		raise "Not implemented"
	end
	
	#
	# Get coordinates of given centroids.
	#
	#@param	i	Index of centroid or indexes described by an Array or Range.
	#
	def getCentroid i
		raise "Not implemented"
	end
	
	#
	# Set position of centroid. Should check dimensionality.
	#
	#@param	i	Index of centroid.
	#@param	pos	New position as array of coordinates. Can also be nil.
	#
	def setCentroid i, pos
		raise "Not implemented"
	end
	
	#
	# Get number of clusters.
	#
	def clusters
		raise "Not implemented"
	end
	
	#
	# Set number of clusters.
	#
	#@param	n	Number of clusters.
	#
	def clusters= n
		raise "Not implemented"
	end
	
	#
	# Saves data and assigned cluster labels in CSV.
	#
	#@param file	Output file.
	#
	def write_csv file, sep = ','
		h = File.open(file, 'w')
		length.times { |i|
			h.write "#{(self[i] + [getLabel(i)]).join sep}\n"
		}
	end
	
	#
	# Draws on given surface.
	#
	#@param	s	Object implementing required drawing methods (set_color, plot_x, plot_o, line, text).
	#@todo		Clear surface first.
	#
	def draw s
		margin = [0.02, 0.02, 0.02, 0.02] #t r b l
		b = boundaries
		
		s.clear
		
		c = -1
#		s.set_hsl_color rand(360), 0, 255
		[0, nil].each { |i|
			s.set_hsl_color 0, 0, (c += 1)*255
			get_points_labeled(i).each { |pid|
				point = self[pid]
				s.plot_o margin[3] + (1 - margin[3] - margin[1]) * (point[0].to_f - b[0][0]) / (b[0][1] - b[0][0]), margin[0] + (1 - margin[0] - margin[2]) * (point[1].to_f - b[1][0]) / (b[1][1] - b[1][0])
			}
		}
		c = -1
		[*1..self.clusters].each { |i|
			s.set_color(c += 1)
			get_points_labeled(i).each { |pid|
				point = self[pid]
				s.plot_o margin[3] + (1 - margin[3] - margin[1]) * (point[0].to_f - b[0][0]) / (b[0][1] - b[0][0]), margin[0] + (1 - margin[0] - margin[2]) * (point[1].to_f - b[1][0]) / (b[1][1] - b[1][0])
			}
		}
		s.set_hsl_color 0, 0, 127
		(get_points_labeled { |l| (not l.nil?) && l < 0 }).each { |pid|
			point = self[pid]
			s.plot_o margin[3] + (1 - margin[3] - margin[1]) * (point[0].to_f - b[0][0]) / (b[0][1] - b[0][0]), margin[0] + (1 - margin[0] - margin[2]) * (point[1].to_f - b[1][0]) / (b[1][1] - b[1][0])
		}
	end
end

