=begin
@author Ville Kumpulainen

A class implementing Dataset "interface". See Dataset for details regarding common methods.

In SimpleDataset data is stored directly to instance variable.

@see	Dataset
=end

%w{dataset metric}.each { |f|
	require f
	#require "./#{f}.rb"
}

#
#@param	data			Stored coordinates of data points.
#@param	dimensionality	All stored vectors must have same dimensionality (as described in this variable).
#@param	mapping			There are some optimizations for which we need to alter the order of data points. However, I do not want to alter data directly so that indexing from an outsider's point of view stays the same. Therefore this variable is used to store indexes of data points in a desired order. This ordering is used internally.
#@param	r_mapping		Reversed mapping: @mapping[@r_mapping[k]] = k.
#
class SimpleDataset < Dataset
	@data = nil
	@dimensionality = nil
	@mapping = nil
	@r_mapping = nil
	
	#
	# Initializer supporting different type of parameters. Parameters can be one of:
	# - Fixnum to set the dimensionality and initialize an empty dataset.
	# - Array consisting of data points,  expressed as another Array representing point coordinates.
	# - File or String referring to a file which contains data.
	#
	#@param	p	Parameters.
	#
	def initialize *p
		raise "Parameter required" unless p.length > 0
		@label = []
		@centroid = []
		@clusters = 0
		case p.length
			when 1
				case p[0].class.to_s
					when 'Fixnum'
						raise "Dimensionality must be a positive integer" unless (@dimensionality = p[0]) > 0
						@data = []
					when 'Array'
						data = p[0] #pass parameter through sanity tests of data=
					when 'File'
						load_data p[0]
					when 'String'
						load_data p[0]
					else
						raise "Unsupported type of a parameter"
				end
			#nothing requiring multiple parameters not implemented
			else
				raise "Unsupported number of parameters"
		end
	end

	def dimensionality
		@dimensionality
	end
	
	def boundaries
		#@todo: store boundaries and update them when new vectors are added
		return @data.inject(@data[0].map{|c|[c]*@dimensionality}) { |b,vector|
			vector.each_with_index {|c,i|
				if c < b[i][0]
					b[i][0] = c
				elsif b[i][1] < c
					b[i][1] = c
				end
			}
			b
		}
	end
	
	def length
		@data.length
	end
	
	#
	# Method for loading data from a file. Each row of a file consists of one data point. On a row each coordinate is separated by spaces.
	#
	#@todo	Check input validity: String::to_f automatically converts invalid strings to 0.0
	#
	def load_data file
		file = file.path if file.class == File
		raise "Given parameter should be instance of String or File" unless file.class == String
		raise "File does not exist" unless File.exists? file
		d = []
		IO.foreach(file) { |line|
			d << line.split.map {|n| n.to_f}
		}
		self.data = d #checks data validity
	end
	
	#
	# Sets array of points directly. Checks data validity. Not included in the base class because passing an array of points (or yet originally storing them to be passed as a parameter) can be inconvenient. Generally an Enumerator based approach is preferable.
	#
	#@param	value	An array consisting of data points, expressed as another Array representing point coordinates.
	#
	def data= value
		raise "Data frozen" if @data.frozen?
		raise "Parameter must be a non-empty two-dimensional array" unless value.class == Array and value.length > 0 and value[0].class == Array
		d = value[0].length
		raise "All points should be #{d}-dimensional" if value.any? {|p| p.length != d}
		raise "Point coordinates should be numerical" unless value.all? {|p| p.all? {|n| [Fixnum, Float].any? {|c| n.class == c}}}
		@dimensionality = d
		@data = value
		@label = Array.new value.length
		forgetMapping
	end
	
	def add value
		raise "Data frozen" if @data.frozen?
		raise "Dimensionality not set" if dim.nil?
		raise "Added point should be #{dim}-dimensional" unless value.class == Array and value.length == dim
		raise "Point coordinates should be numerical" unless value.all? {|n| [Fixnum, Float].any? {|c| n.class == c}}
		@data << value
		@label << nil
		forgetMapping
	end
	
	def [] i
		@data[i]
	end
	
	#
	# When data is altered, mapping used for optimizations in queries must be forgotten.
	#
	def forgetMapping
		@mapping = nil
	end
	
	#
	# Updates data mapping used in query optimizations. Variable @mapping preserves indexes pointing to @data ordered ascendingly by first coordinate.
	#
	def updateMapping
		if @mapping.nil?
			#@todo	Sort by axis having the widest distribution (..or even by a line (..discovered with PCA? (Should this really be considered to be implemented in a class called "SimpleDataset"? (Way too many recursive questions?))))
			@mapping = [*0...@data.length].sort_by {|i| @data[i][0]}
			@r_mapping = Array.new @mapping.length
			@mapping.each_with_index {|d,i| @r_mapping[d] = i}
		end
	end
	
	private :forgetMapping, :updateMapping
	
	def test_mapping
		raise "A" if [*0...@data.length].any? {|i| @mapping[@r_mapping[i]] != i}
		raise "B" if [*0...@data.length].any? {|i| @r_mapping[@mapping[i]] != i}
	end

=begin
For the record, this is difference of two indexing systems:
presentation layer		sorted (optimized) layer
		a					@r_mapping[a]
	@mapping[a]					a

@todo: any improvements to the naming of these two associated arrays are welcome :)
=end

	def regionalQuery i, r
		updateMapping
		neighbors = []
		j = 0
		# Keep looking for sorted (possible) neighbors, until..
		while j += 1 do
			pts = [] #pts contains candidates for nearest neighbors
			[@mapping[@r_mapping[i] - j], @mapping[@r_mapping[i] + j]].each {|index|
				pts << index if (0...@data.length).include?(index)
			}
			
			pts.each { |p|
				if block_given?
					neighbors << p if yield(@data[i], @data[p])
				else
					neighbors << p if Metric::distance(@data[i], @data[p]) <= r
				end
			}
			# ..distance between considered points and i(th) point along x-axis grows too long.
			break if pts.empty? or pts.all? { |p| (@data[i][0] - @data[p][0]).abs > r }
		end
		return neighbors
	end

	def nearestNeighbors i, k
		raise "k must be a Fixnum" unless k.class == Fixnum
		raise "Not enough data" if k > @data.length

		updateMapping
		neighbors = [] #We'll store index-distance pairs in this array (which reduces repeating distance calculations).
		j = 0
		while neighbors.length < k do #Fill neighbors with initial candidates.
			j += 1
			[@mapping[@r_mapping[i] - j], @mapping[@r_mapping[i] + j]].each {|index|
				neighbors << [index, Metric::distance(@data[i], @data[index])] if (0...@data.length).include?(index)
			}
		end
		neighbors.sort_by! {|n| n[1]}
		neighbors.pop if neighbors.length > k #Neighbor candidates have been added pairwise, hence there can be one too many.
		
		while j += 1 do #always true
			pts = [] #pts contains candidates for nearest neighbors
			[@mapping[@r_mapping[i] - j], @mapping[@r_mapping[i] + j]].each {|index|
				pts << [index, Metric::distance(@data[i], @data[index])] if (0...@data.length).include?(index)
			}
			
			#Leave if distance to neighbors.last is less than distance along x-axis to upcoming points.
			break if pts.all? {|p| neighbors.last[1] < (@data[p[0]][0] - @data[i][0]).abs }
			
			pts.each { |p|
				if p[1] <= neighbors.last[1]
					neighbors << p
					neighbors.sort_by! {|n| n[1]} #@todo: Make use of knowledge that neighbors was originally in order (=> time complexity of this portion reduces to ~O(n)!)
					neighbors.pop
				end
			}
		end
		return neighbors.map {|n| n[0]} #Pick indexes to an array.
	end
	
	def lockData
		#@todo: Disable access to methods modifying @data.
		@data.freeze
	end
	
	private :lockData
	
	def getLabel i
		raise "Variable i is not a Fixnum" unless i.class == Fixnum
		@label[i]
	end
	
	def setLabel i, l
		raise "Variable i is not a Fixnum" unless i.class == Fixnum
		raise "Variable l is not a Fixnum or nil" unless l.nil? || l.class == Fixnum
		#@todo: other sanity checks? affects dbscan..
		lockData
		@label[i] = l
	end
	
	def get_points_labeled label = 0
		returned = []
		@label.each_with_index { |l,i|
			if block_given?
				returned << i if yield(l)
			else
				returned << i if label == l
			end
		}
#		puts returned.inspect
		return returned
	end
	
	def getCentroid i
		raise "Variable i is not a Fixnum" unless i.class == Fixnum
		@centroid[i]
	end
	
	def setCentroid i, pos
		raise "Variable i is not a Fixnum" unless i.class == Fixnum
		raise "Variable pos must be either a #{dim}-dimensional array of Floats/Fixnums or nil" unless pos.nil? || ( pos.class == Array && pos.length == dim && pos.all?{|p| [Fixnum, Float].any?{|c| p.class == c}} )
		lockData
		@centroid[i] = pos
	end
	
	def clusters
		@clusters
	end
	
	def clusters= n
		raise "Variable n is not a non-negative Fixnum" unless n.class == Fixnum && n >= 0
		lockData
		(@clusters - n).times { @centroid.pop }
		(n - @clusters).times { @centroid.push nil }
		@label.each_with_index { |l,i| @label[i] = nil if l >= n } if @clusters > n
		@clusters = n
	end
	
	def debug
		puts "label: #{@label.length}x1"
		puts "clusters: #{@clusters}"
	end
end
