#!/usr/bin/ruby
require './include/add_path.rb'

require 'clustering'
require 'simple_dataset'
require 'drawable_sdl'

def alert msg
	puts "**#{'*' * msg.length}**"
	puts "* #{' ' * msg.length} *"
	puts "* #{      msg       } *"
	puts "* #{' ' * msg.length} *"
	puts "**#{'*' * msg.length}**"
end

def help long = false
	puts "Usage: #{$0} ((input.txt | -c out.csv | -o output.bmp | -p dbscan_parameter value)* | (-n | -l log.txt | -b | -h | -s 800 600))"
	puts "Keys: esc - abort, space - next (once clustering has finished)"
	puts "See #{$0} -h for details." if not long
	return unless long
	puts
	{
		'-h' => 'Displays this help.',
		'-l' => 'Log file (usage: -l log.txt).',
		'-b' => 'Run in batch mode: never wait for user input between clusterings.',
		'-s' => 'Set canvas size for drawing (usage: -s 800 600).',
		'-o' => 'Output image file for the clustering (usage: -o out.bmp).',
		'-c' => 'Output csv file for the clustering (usage: -c out.csv)',
		'-n' => 'Headless mode. Outputting images is disabled and batch mode is set on.',
		'-p' => 'Pass optional parameters to DBSCAN (usage: -p splits 3).',
		'-pr'=> 'Sets priority of clustering thread (usage: -pr 65535). Use negative values with consideration!'
	}.each_pair { |k,v| puts "\t#{k}\t#{v}" }
	puts
	puts "Number of output files must either be equal to.."
	puts "\t0:\tNo output to image files."
	puts "\t1:\tUse given file name as a rule to generate output files. Given output file MUST contain one of replacables (see below)."
	puts "\t..number of input files:"
	puts "\t  \tEach clustering derived from input file is written to output file. Input and output files are linked together corresponding to order of appearance on command line."
	puts
	puts "Replacables for output file (image or csv) name:"
	puts
	{
		'%n' => 'Base name of the input file without root path and file extension.',
		'%i' => 'Number of the file processed. Numbers are assigned by their order in program parameters.'
	}.each_pair { |k,v| puts "\t#{k}\t#{v}" }
end

ifiles = []
cfiles = []
ofiles = []
logfile = nil
dbscan_parameters = {}
batch_mode = false
size = [800, 600]
priority = 256*256 - 1
no_display = false

starting_time = Time.now.to_f

begin #Parse command line arguments.
	i = 0
	while i < ARGV.length do
		case (a = ARGV[i])
			when %r{^-n$}
				no_display = true
				batch_mode = true

			when %r{^-+(h(elp)?|\?.*)$} #-h, --help, -? (and other varieties..)
				help true
				exit

			when %r{^--?c(sv)?$}
				raise "#{a} should be followed by a parameter!" unless (i += 1) < ARGV.length
				cfiles << ARGV[i]
		
			when %r{^--?o(utput(-?file)?)?$}
				raise "#{a} should be followed by a parameter!" unless (i += 1) < ARGV.length
				ofiles << ARGV[i]
			
			when %r{^--?l(og(-?file)?)?$}
				raise "#{a} should be followed by a parameter!" unless (i += 1) < ARGV.length
				logfile = ARGV[i]
		
			when %r{^--?p(aram(eter)?)?$}
				raise "#{a} should be followed by two parameters (parameter, value)!" unless (i += 2) < ARGV.length
				dbscan_parameters[ARGV[i-1]] = ARGV[i].to_i #so far parameters for DBSCAN are numerical.. note: invalid parameters are converted to 0
				
			when %r{^--?pr(iority)?$}
				raise "#{a} should be followed by a parameter!" unless (i += 1) < ARGV.length
				priority = ARGV[i].to_i #note: invalid parameters are converted to 0
		
			when %r{^--?b(atch(-?mode)?)?$}
				batch_mode = true
				
			when %r{^--?s(ize)?$}
				raise "#{a} should be followed by two parameters (width, height)!" unless (i += 2) < ARGV.length
				size = ARGV[(i-1)..i].map{|i|i.to_i}
				raise "Given measures should be at least 50." if size.any? {|s|s<50}
		
			else
				raise "File #{a} not found!" unless File.exists? a
				ifiles << a
		end
		i += 1
	end

	raise "No input file specified!" if ifiles.length == 0

	case ofiles.length
		when 0 #OK (we simply don't write images at all (@todo a small delay or user input?))
		when 1
			raise "Output file must contain replacable substring (or output files must be individually named for all input files)." if 1 < ifiles.length && %w{%n %i}.all? { |r| ofiles[0][r].nil? }
		when ifiles.length #OK as well
		else
			raise "Number of input files (#{ifiles.length}) and output files (#{ofiles.length}) does not match!"
	end

	#same applies for csv files
	case cfiles.length
                when 0 #OK (we simply don't write csv files at all
                when 1
                        raise "Output file must contain replacable substring (or output files must be individually named for all input files)." if 1 < ifiles.length && %w{%n %i}.all? { |r| ofiles[0][r].nil? }
                when ifiles.length #OK as well
                else
                        raise "Number of input files (#{ifiles.length}) and output files (#{cfiles.length}) does not match!"
        end


rescue Exception => ex #Clean exit :)
	puts ex
	help unless ex.class == SystemExit
	exit
end

if not logfile.nil?
	lh = File.open(logfile, 'a')
	lh.write "--- Log opened at #{Time.now}\n\n"
	if dbscan_parameters.empty?
		lh.write "Starting #{ifiles.length} DBSCAN runs with default parameters.\n"
	else
		lh.write "Starting #{ifiles.length} DBSCAN runs with following parameters:\n"
		dbscan_parameters.each_pair { |k,v|
			lh.write "#{k} := #{v}\n"
		}
		lh.write "\n"
	end
	lh.write "Input\tSections / Splits\tClusters\tIsolated\tEps\tInit time\tCl. time\n"
	lh.flush
end

d = DrawableSDLSurface.new size[0], size[1] unless no_display

abort = false

ifiles.each_with_index { |file, i|
	last = (i == ifiles.length - 1)
	
	#load data
	s = SimpleDataset.new file
	
	#initialize
	alert "Initializing DBSCAN for #{file}"
	time = Time.now.to_f
	#@see DBSCAN for more details on possible parameters.
	c = Clustering.create s, Clustering::Algorithm::DBSCAN, dbscan_parameters
	initialization_time = Time.now.to_f - time
	
	#start clustering
	alert "Starting clustering for #{file}"
	clustering_time = 0
	(t = Thread.new {
		time = Time.now.to_f
		while not c.happy? do c.step end
		clustering_time = Time.now.to_f - time
	}).priority = priority

	running = true
	event_thread = Thread.new {
		while not no_display
			while event = SDL::Event2.poll
				case event
					when SDL::Event2::KeyDown
						case event.sym
							when 27 #esc
								abort = not(last and c.happy?)
								running = false
					
							when 32 #space
								running = (not c.happy?) #next please
						
							else
								puts "Unknown key press #{event.sym} (supported keys are space and esc)"
						end
			
					when SDL::Event2::Quit
						abort = not(last and c.happy?)
						running = false
				end
			end
		end
	}
	
	event_thread.priority = priority if priority > 0
	
	while running
		#show some output to the user occasionally
		unless no_display
			c.data.draw d
			d.display unless no_display
		end
		if(batch_mode && (running = (not c.happy?)))
			sleep 1.5 #spare some cpu, we don't need to provide real-time information to the user
		end
	end
	
	Thread.kill event_thread
	
	t.join #wait until clustering is truly finished
	
	unless no_display
		c.data.draw d
		d.display
	end

	if ofiles.length > 0
		img_file = ofiles[ofiles.length > 1 ? i : 0].dup
		temp = file.split('/').pop.split('.') # @todo take windows path in consideration ("C:\ruby\..")
		temp.pop		
		{
			'%i' => "%0#{Math.log(ifiles.length,10).to_i+1}d" % (i+1), #logarithm makes sure there is enough leading zeroes
			'%n' => temp.join('.')
		}.each_pair { |k,v| img_file.gsub! k, v }
		d.write(img_file) unless no_display
	end

	#yet again, same for csv files
	if cfiles.length > 0
		csv_file = cfiles[cfiles.length > 1 ? i : 0].dup
		temp = file.split('/').pop.split('.') # @todo take windows path in consideration ("C:\ruby\..")
		temp.pop
		{
			'%i' => "%0#{Math.log(ifiles.length,10).to_i+1}d" % (i+1),
			'%n' => temp.join('.')
		}.each_pair { |k,v| csv_file.gsub! k, v }
		c.data.write_csv(csv_file)
	end

	if not logfile.nil?	
		lh.write ((%w{%s %\ 4d\ /%\ 4d %\ 4d %\ 4d %6.3f %6.3f %6.3f}.join("\t") + "\n") % [file, c.sections, c.splits, c.data.clusters, c.data.get_points_labeled(0).length, c.threshold, initialization_time, clustering_time])
		lh.flush
	end
	
	break if abort
}

puts "Clustering aborted!" if abort
puts "Total running time: #{Time.now.to_i - starting_time.to_i} seconds"

if not logfile.nil?
	lh.write "Clustering aborted!\n" if abort
	lh.write "\nTotal running time: #{Time.now.to_i - starting_time.to_i} seconds\n\n" #integer is probably better for this usage
	lh.write "--- Log closed at #{Time.now}\n"
	lh.close
end

